# flake8: noqa

from .main import *
from . import ops
from . import sim
from . import utils
from . import help
from . import demos
from .main import *

__docformat__ = 'restructuredtext'

#package-level docstring
__doc__ = """
Geonomics - A Python package for running complex landscape genomics simulations
===============================================================================

"""
#with open('../README.rst', 'r') as f:
#    __doc__ = __doc__ + f.read()

